from pprint import pformat
from collections import OrderedDict
from datetime import datetime

try:
    from UserDict import DictMixin
except ImportError:
    from collections import MutableMapping as DictMixin

try:
    text = unicode
except NameError:
    text = str 


# Use the same json implementation as itsdangerous because Flask does that
# and we depend on it
try:                                                                               
    from itsdangerous import simplejson as _json                                
except ImportError:                                                             
    from itsdangerous import json as _json                                      


def convert_column_name(column_name):
    return column_name.lower().replace(' ','_')


class TupleMeta(object):

    def __init__(self, input_data=None, select_vector=None, **kw_args):
        buf = {}

        if input_data is None:
            input_data = kw_args

        elif isinstance(input_data, list):
            input_data = dict(input_data)

        elif not (isinstance(input_data, dict) or isinstance(input_data, TupleMeta)):
            raise ValueError('input must be either of type dict or list')

        if select_vector is not None and len(select_vector) != 0:

            if len(select_vector) > len(self.__slots__):
                raise ValueError('Source select vector is too long')
        else:
            select_vector = self.__slots__

        for k,v in input_data.items():
            buf[convert_column_name(k)] = v 
        
        for i, slot_key in enumerate(self.__slots__):

            if i < len(select_vector):
                setattr(self, slot_key, buf.get(select_vector[i]))
            else:
                setattr(self, slot_key, buf.get(slot_key))

    def __getitem__(self, k):
        if isinstance(k, int):
            return getattr(self,self.__slots__[k])

        if k in self.__slots__:
            return getattr(self, k)

    def __contains__(self, value):
        return value in self.__slots__

    def __eq__(self, value):
        if len(self.__slots__) == 1:
            stored_value = self[0]

            if isinstance(value, type(stored_value)) \
                or (isinstance(stored_value, text) and isinstance(value, str)) \
                or (isinstance(stored_value, str) and isinstance(value, text)):

                return stored_value == value
       
        for i, k in enumerate(self.__slots__):
            if getattr(self, k) != value[i]:
                return False

        return True

    def __ne__(self, value):

        for k in self.__slots__:
            if getattr(self, k) != value[k]:
                return True

        return False

    def __repr__(self):
        d = self.items()
        return 'Tuple %s' % d.__repr__()

    def __hash__(self):

        if len(self.__slots__) == 1:
            return hash(self.values()[0])

        return hash(tuple(self.values()))

    def __len__(self):
        return len(self.__slots__)

    def get(self, k, default=None):
        if k in self.__slots__:

            try:
                return getattr(self, k)

            except AttributeError:
                return default

        return None

    def keys(self):
        return self.__slots__

    def items(self):
        result = []
        for k in self.__slots__:
            result.append((k, getattr(self, k)))

        return result

    def values(self):
        result = []
        for k in self.__slots__:
            result.append(getattr(self, k))

        return result

    def select(self, keys):
        result = {}

        for k in keys:
            if k in self.__slots__:
                result[k] = getattr(self, k)

        return result

    def __dict__(self):
        return self.asdict()

    def asdict(self):
        result = {}

        for k in self.__slots__:
            result[k] = getattr(self, k) 

        return result

    def asjson(self):
        return _json.dumps(self.asdict(), cls=TupleJSONEncoder)


def build_tuple_type(*columns):

    if len(columns) == 1 and isinstance(columns[0], list):
        columns = columns[0]

    columns = [convert_column_name(c) for c in columns]

    class Tuple(TupleMeta): 
        __slots__ = columns

    return Tuple


class TupleJSONEncoder(_json.JSONEncoder):
    
    def default(self, obj):

        if isinstance(obj, datetime):
            return obj.isoformat()

        asdict_op = getattr(obj, "asdict", None)
        if callable(asdict_op):
            return obj.asdict()

        else:
            return _json.JSONEncoder.default(self, obj)


class BaseItem(object):
    """Base class for all scraped items."""
    pass


class Field(dict):
    """Container of field metadata"""
    pass


class ItemMeta(type):

    def __new__(mcs, class_name, bases, attrs):
        fields = OrderedDict()
        new_attrs = OrderedDict()
        for n, v in attrs.iteritems():
            if isinstance(v, Field):
                fields[n] = v
            else:
                new_attrs[n] = v

        cls = super(ItemMeta, mcs).__new__(mcs, class_name, bases, new_attrs)
        cls.fields = cls.fields.copy()
        cls.fields.update(fields)
        return cls


class DictItem(DictMixin, BaseItem):

    fields = OrderedDict()

    def __init__(self, *args, **kwargs):
        self._values = OrderedDict()
        if args or kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).iteritems():
                self[k] = v

    def __getitem__(self, key):
        return self._values[key]

    def __setitem__(self, key, value):
        if key in self.fields:
            self._values[key] = value
        else:
            raise KeyError("%s does not support field: %s" %
                (self.__class__.__name__, key))

    def __delitem__(self, key):
        del self._values[key]

    def __getattr__(self, name):
        if name in self.fields:
            raise AttributeError("Use item[%r] to get field value" % name)
        raise AttributeError(name)

    def __setattr__(self, name, value):
        if not name.startswith('_'):
            raise AttributeError("Use item[%r] = %r to set field value" %
                (name, value))
        super(DictItem, self).__setattr__(name, value)

    def keys(self):
        return self._values.keys()

    def __repr__(self):
        return pformat(dict(self))

    def copy(self):
        return self.__class__(self)

    def __len__(self):
        return len(self.fields)

    def __iter__(self):
        return self

    def load(self, source):
        
        for key in self.fields.keys():

            value = source.get(key)

            if value is None:
                self[key] = None

            elif isinstance(value, str):
                self[key] = value

            elif isinstance(value, text):
                self[key] = value.encode('utf-8', 'ignore')
            
            elif isinstance(value, list):
                self[key] = _json.dumps(value)

            elif isinstance(value, dict):
                self[key] = _json.dumps(value)
            
            else:
                self[key] = value


    def select(self, keys):
        result = []

        for key in keys:
            if key in self.keys():
                result.append(self[key])

        return result

    def add_field(self, name, **kwargs):
        field = Field(**kwargs)
        self.fields[name] = field


class Item(DictItem):

    __metaclass__ = ItemMeta


class seekable(object):
    
    def __init__(self, iterable):
        self.pos = 0
        self.buf = []
        self.buf_iterator = iter(self.buf)
        self.disabled = False

    def __iter__(self):
        return self

    def __next__(self):
        self.disabled = True

        if self.pos < len(self.buf):
            b=self.buf[self.pos]
            self.pos+=1
            return b
        else:
            return next(self.iterator)

    def __getitem__(self, index):

        # getting an item is still posible but we are rather raising error
        # to prevent incorrect usage
        if not self.disabled:

            if index < len(self.buf):
                return self.buf[index]
            else:
                self.seek(index)
                return self.buf[index]

        else:
            raise ValueError('Seeking disabled iteration already started')

    def seek(self, index):

        if not self.disabled:

            for i in range(index-(len(self.buf)-1)):
                self.buf.append(next(self.iterator))
                    
        else:
            raise ValueError('Seeking disabled iteration already started')

    def is_empty(self):
        try:
            self.seek(1)
            return False

        except StopIteration:
            return True
     
    def has_one_element(self):
        test_first = False
        test_second = False

        try:
            self.seek(1)
            test_first = True

        except StopIteration:
            test_first = False

        try:
            self.seek(2)
            test_first = False

        except StopIteration:
            test_first = True

        return test_first and not test_second

    def get_buffer_iterator(self):
        return iter(self.buf)

